from typing import Tuple

import numpy as np
import scipy.special as spc
import scipy.stats as st

from ..base import Distribution, LinkFunction, ScipyMixin
from ..links import Log


class Gamma(ScipyMixin, Distribution):
    """The Gamma Distribution for GAMLSS.

    The distribution function is defined as in GAMLSS as:
    $$
    f(y|\mu,\sigma)=\\frac{y^{(1/\sigma^2-1)}\exp[-y/(\sigma^2 \mu)]}{(\sigma^2 \mu)^{(1/\sigma^2)} \Gamma(1/\sigma^2)}
    $$

    with the location and shape parameters $\mu, \sigma > 0$.

    !!! Note
        The function is parameterized as GAMLSS' GA() distribution.

        This parameterization is different to the `scipy.stats.gamma(alpha, loc, scale)` parameterization.

        We can use `Gamma().theta_to_scipy_params(theta)` to map the distribution parameters to scipy.

    The `scipy.stats.gamma()` distribution is defined as:
    $$
    f(x, \\alpha, \\beta) = \\frac{\\beta^\\alpha x^{\\alpha - 1} \exp[-\\beta x]}{\Gamma(\\alpha)}
    $$
    with the paramters $\\alpha, \\beta >0$. The parameters can be mapped as follows:
    $$
    \\alpha = 1/\sigma^2 \Leftrightarrow \sigma = \sqrt{1 / \\alpha}
    $$
    and
    $$
    \\beta = 1/(\sigma^2\mu).
    $$

    Args:
        loc_link (LinkFunction, optional): The link function for $\mu$. Defaults to Log().
        scale_link (LinkFunction, optional): The link function for $\sigma$. Defaults to Log().
    """

    corresponding_gamlss: str = "GA"

    parameter_names = {0: "mu", 1: "sigma"}
    parameter_support = {
        0: (np.nextafter(0, 1), np.inf),
        1: (np.nextafter(0, 1), np.inf),
    }
    distribution_support = (np.nextafter(0, 1), np.inf)
    # Scipy equivalent and parameter mapping ondil -> scipy
    scipy_dist = st.gamma
    # Theta columns do not map 1:1 to scipy parameters for gamma
    # So we have to overload theta_to_scipy_params
    scipy_names = {}

    def __init__(
        self,
        loc_link: LinkFunction = Log(),
        scale_link: LinkFunction = Log(),
    ) -> None:
        super().__init__(links={0: loc_link, 1: scale_link})

    def theta_to_scipy_params(self, theta: np.ndarray) -> dict:
        """Map GAMLSS Parameters to scipy parameters.

        Args:
            theta (np.ndarray): parameters

        Returns:
            dict: Dict of (a, loc, scale) for scipy.stats.gamma(a, loc, scale)
        """
        mu = theta[:, 0]
        sigma = theta[:, 1]
        beta = 1 / (sigma**2 * mu)
        params = {"a": 1 / sigma**2, "loc": 0, "scale": 1 / beta}
        return params

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma = self.theta_to_params(theta)

        if param == 0:
            return (y - mu) / ((sigma**2) * (mu**2))

        if param == 1:
            return (2 / sigma**3) * (
                (y / mu)
                - np.log(y)
                + np.log(mu)
                + np.log(sigma**2)
                - 1
                + spc.digamma(1 / (sigma**2))
            )

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma = self.theta_to_params(theta)
        if param == 0:
            # MU
            return -1 / ((sigma**2) * (mu**2))

        if param == 1:
            # SIGMA
            return (4 / sigma**4) - (4 / sigma**6) * spc.polygamma(1, (1 / sigma**2))

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        if sorted(params) == [0, 1]:
            return np.zeros_like(y)

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        initial_params = [np.mean(y), 1]
        return np.tile(initial_params, (y.shape[0], 1))



import numba as nb
import numpy as np

SMALL_NUMBER = 1e-10
LARGE_NUMBER = 1e15
LOG_SMALL_NUMBER = np.log(SMALL_NUMBER)
LOG_LARGE_NUMBER = np.log(LARGE_NUMBER)


@nb.vectorize(["float64(float64)", "float32(float32)"])
def robust_log(x: np.ndarray) -> np.ndarray:
    """
    A robust log function that handles negative and zero values.

    This function returns the logarithm of the input array, replacing
    negative and zero values with a small positive number to avoid
    undefined logarithm values.
    """
    if x > SMALL_NUMBER:
        return np.log(x)
    else:
        return LOG_SMALL_NUMBER


@nb.vectorize(["float64(float64)", "float32(float32)"])
def robust_exp(x: np.ndarray) -> np.ndarray:
    """
    A robust exponential function that handles large values.
    """

    if x > LOG_LARGE_NUMBER:
        return LARGE_NUMBER
    else:
        return np.exp(x)


@nb.vectorize(["float32(float32, float32)", "float64(float64, float64)"])
def robust_sigmoid(value, k):
    if value / k > 50:
        return 1
    elif value / k < -50:
        return 0
    else:
        return 1 / (1 + robust_exp(-k * (value - 1)))


@nb.vectorize(["float32(float32)", "float64(float64)"])
def robust_softplus(value):
    if value > 40:
        return value
    else:
        return np.log(1 + np.exp(value))


@nb.vectorize(["float32(float32)", "float64(float64)"])
def robust_softplus_inverse(value):
    if value > 40:
        return value
    elif value < SMALL_NUMBER:
        return -30
    else:
        return np.log(np.exp(value) - 1)


@nb.vectorize(["float32(float32, float32)", "float64(float64, float64)"])
def zero_safe_division(a, b):
    """Return the max float value for the current precision at the a / b if
    b becomes 0. Returns

    Args:
        a (float): Nominator.
        b (float): Denominator.

    Returns:
        float: Result of the division a / b.
    """
    if np.isclose(b, 0):
        return LARGE_NUMBER
    elif np.isclose(a, 0):
        return LARGE_NUMBER
    else:
        return a / b



from typing import Tuple

import numpy as np
import scipy.special as spc
import scipy.stats as st

from ..base import Distribution, LinkFunction, ScipyMixin
from ..links import Logit


class Beta(ScipyMixin, Distribution):
    """The Beta Distribution for GAMLSS.

    The distribution function is defined as in GAMLSS as:
    $$
    f(y|\\mu,\\sigma)=\\frac{\\Gamma(\\frac{1 - \sigma^2}{\sigma^2})}
        {
        \\Gamma(\\frac{\\mu (1 - \\sigma^2)}{\\sigma^2})
        \Gamma(\\frac{(1 - \\mu) (1 - \sigma^2)}{\\sigma^2})}
        y^{\\frac{\mu (1 - \\sigma^2)}{\\sigma^2} - 1}
        (1-y)^{\\frac{(1 - \\mu) (1 - \\sigma^2)}{\\sigma^2} - 1}
    $$



    with the location and shape parameters $\mu, \sigma > 0$.

    !!! Note
        The function is parameterized as GAMLSS' BE() distribution.

        This parameterization is different to the `scipy.stats.beta(alpha, beta, loc, scale)` parameterization.

        We can use `Beta().gamlss_to_scipy(mu, sigma)` to map the distribution parameters to scipy.

    The `scipy.stats.beta()` distribution is defined as:
    $$
    f(x, \\alpha, \\beta) = \\frac{\\Gamma(\\alpha + \\beta) x^{\\alpha - 1} {(1 - x)}^{\\beta - 1}}{\Gamma(\\alpha) \\Gamma(\\beta)}
    $$

    with the paramters $\\alpha, \\beta >0$. The parameters can be mapped as follows:
    $$
    \\alpha = \\mu (1 - \\sigma^2) / \\sigma^2 \\Leftrightarrow \\mu = \\alpha / (\\alpha + \\beta)
    $$
    and
    $$
    \\beta = (1 - \\mu) (1 - \\sigma^2)/ \\sigma^2 \\Leftrightarrow \\sigma = \\sqrt{((\\alpha + \\beta + 1) )}
    $$


    Args:
        loc_link (LinkFunction, optional): The link function for $\mu$. Defaults to  LOGIT
        scale_link (LinkFunction, optional): The link function for $\sigma$. Defaults to LOGIT
    """

    corresponding_gamlss: str = "BE"

    parameter_names = {0: "mu", 1: "sigma"}
    parameter_support = {
        0: (np.nextafter(0, 1), np.nextafter(1, 0)),
        1: (np.nextafter(0, 1), np.nextafter(1, 0)),
    }
    distribution_support = (np.nextafter(0, 1), np.nextafter(1, 0))
    # Scipy equivalent and parameter mapping ondil -> scipy
    scipy_dist = st.beta
    # Theta columns do not map 1:1 to scipy parameters for beta
    # So we have to overload theta_to_scipy_params
    scipy_names = {}

    def __init__(
        self,
        loc_link: LinkFunction = Logit(),
        scale_link: LinkFunction = Logit(),
    ) -> None:
        super().__init__(links={0: loc_link, 1: scale_link})

    def theta_to_scipy_params(self, theta: np.ndarray) -> dict:
        """Map GAMLSS Parameters to scipy parameters.

        Args:
            theta (np.ndarray): parameters

        Returns:
            dict: Dict of (a, b, loc, scale) for scipy.stats.beta(a, b, loc, scale)
        """
        mu = theta[:, 0]
        sigma = theta[:, 1]
        alpha = mu * (1 - sigma**2) / sigma**2
        beta = (1 - mu) * (1 - sigma**2) / sigma**2
        params = {"a": alpha, "b": beta, "loc": 0, "scale": 1}
        return params

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma = self.theta_to_params(theta)

        if param == 0:
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            return ((1 - sigma**2) / sigma**2) * (
                -spc.digamma(alpha) + spc.digamma(beta) + np.log(y) - np.log(1 - y)
            )

        if param == 1:
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            return -(2 / sigma**3) * (
                mu * (-spc.digamma(alpha) + spc.digamma(alpha + beta) + np.log(y))
                + (1 - mu)
                * (-spc.digamma(beta) + spc.digamma(alpha + beta) + np.log(1 - y))
            )

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma = self.theta_to_params(theta)
        if param == 0:
            # MU
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            return -(((1 - sigma**2) ** 2) / sigma**4) * (
                spc.polygamma(1, alpha) + spc.polygamma(1, beta)
            )

        if param == 1:
            # SIGMA
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            return -(4 / (sigma**6)) * (
                (mu**2) * spc.polygamma(1, alpha)
                + ((1 - mu) ** 2) * spc.polygamma(1, beta)
                - spc.polygamma(1, alpha + beta)
            )

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        mu, sigma = self.theta_to_params(theta)

        if sorted(params) == [0, 1]:
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            return (2 * (1 - sigma**2) / sigma**5) * (
                mu * spc.polygamma(1, alpha) - (1 - mu) * spc.polygamma(1, beta)
            )

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        initial_params = [np.mean(y, axis=0), 0.5]
        return np.tile(initial_params, (y.shape[0], 1))



from typing import Tuple

import numpy as np
import scipy.stats as st

from ..base import Distribution, LinkFunction, ScipyMixin
from ..links import Identity, Log


class Normal(ScipyMixin, Distribution):
    """
    The Normal distribution with mean and standard deviation parameterization.

    The probability density function of the distribution is defined as:
    $$
        f(y | \\mu, \\sigma) = \\frac{1}{\\sqrt{2\\pi\\sigma^2}} \exp\\left(-\\frac{(y - \\mu)^2}{2\\sigma^2}\\right).
    $$
    respectively
    $$
        f(y | \\theta_0, \\theta_1) = \\frac{1}{\\sqrt{2\\pi\\theta_1^2}} \exp\\left(-\\frac{(y - \\theta_0)^2}{2\\theta_1^2}\\right).
    $$
    where $y$ is the observed data, $\\mu = \\theta_0$ is the location parameter and $\\sigma = \\theta_1$ is the scale parameter.

    This distribution corresponds to the NO() distribution in GAMLSS.
    """

    corresponding_gamlss: str = "NO"
    parameter_names = {0: "mu", 1: "sigma"}
    parameter_support = {0: (-np.inf, np.inf), 1: (np.nextafter(0, 1), np.inf)}
    distribution_support = (-np.inf, np.inf)

    # Scipy equivalent and parameter mapping ondil -> scipy
    scipy_dist = st.norm
    scipy_names = {"mu": "loc", "sigma": "scale"}

    def __init__(
        self,
        loc_link: LinkFunction = Identity(),
        scale_link: LinkFunction = Log(),
    ) -> None:
        """Initialize the Normal.

        Args:
            loc_link (LinkFunction, optional): Location link. Defaults to Identity().
            scale_link (LinkFunction, optional): Scale link. Defaults to Log().
        """
        super().__init__(
            links={
                0: loc_link,
                1: scale_link,
            }
        )

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma = self.theta_to_params(theta)

        if param == 0:
            return (1 / sigma**2) * (y - mu)

        if param == 1:
            return ((y - mu) ** 2 - sigma**2) / (sigma**3)

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        _, sigma = self.theta_to_params(theta)
        if param == 0:
            # MU
            return -(1 / sigma**2)

        if param == 1:
            # SIGMA
            return -(2 / (sigma**2))

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        if sorted(params) == [0, 1]:
            return np.zeros_like(y)

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        initial_params = [np.mean(y), np.std(y, ddof=1)]
        return np.tile(initial_params, (y.shape[0], 1))


class NormalMeanVariance(ScipyMixin, Distribution):
    """
    The Normal distribution with mean and variance parameterization.

    The probability density function of the distribution is defined as:
    $$
        f(y | \\mu, \\sigma^2) = \\frac{1}{\\sqrt{2\\pi\\sigma^2}} \exp\\left(-\\frac{(y - \\mu)^2}{2\\sigma^2}\\right).
    $$
    respectively
    $$
        f(y | \\theta_0, \\theta_1) = \\frac{1}{\\sqrt{2\\pi\\theta_1}} \exp\\left(-\\frac{(y - \\theta_0)^2}{2\\theta_1}\\right).
    $$
    where $y$ is the observed data, $\\mu = \\theta_0$ is the location parameter and $\\sigma^2 = \\theta_1$ is the scale parameter.
    """

    corresponding_gamlss: str = "NO2"
    parameter_names = {0: "mu", 1: "sigma"}
    parameter_support = {0: (-np.inf, np.inf), 1: (np.nextafter(0, 1), np.inf)}
    distribution_support = (-np.inf, np.inf)

    # Scipy equivalent and parameter mapping ondil -> scipy
    scipy_dist = st.norm
    scipy_names = {"mu": "loc", "sigma": "scale"}

    def __init__(
        self,
        loc_link: LinkFunction = Identity(),
        scale_link: LinkFunction = Log(),
    ) -> None:
        """Initialize the NormalMeanVariance.

        Args:
            loc_link (LinkFunction, optional): Location link. Defaults to Identity().
            scale_link (LinkFunction, optional): Scale link. Defaults to Log().
        """
        super().__init__(
            links={
                0: loc_link,
                1: scale_link,
            }
        )

    def theta_to_scipy_params(self, theta: np.ndarray) -> dict:
        """Map GAMLSS Parameters to scipy parameters.

        Args:
            theta (np.ndarray): parameters

        Returns:
            dict: Dict of (loc, scale) for scipy.stats.norm(loc, scale)
        """
        mu = theta[:, 0]
        sigma = theta[:, 1]
        params = {"loc": mu, "scale": sigma**0.5}
        return params

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma = self.theta_to_params(theta)

        if param == 0:
            return (1 / sigma) * (y - mu)

        if param == 1:
            return 0.5 * ((y - mu) ** 2 - sigma) / (sigma**2)

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        _, sigma = self.theta_to_params(theta)
        if param == 0:
            # MU
            return -(1 / sigma)

        if param == 1:
            # SIGMA
            return -(1 / (2 * sigma**2))

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        if sorted(params) == [0, 1]:
            return np.zeros_like(y)

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        initial_params = [np.mean(y), np.var(y, ddof=1)]
        return np.tile(initial_params, (y.shape[0], 1))



# Author: Simon Hirsch
# Licence: MIT

import numpy as np
import pytest

from ondil.links import (
    Identity,
    InverseSoftPlus,
    InverseSoftPlusShiftValue,
    LogIdent,
    Logit,
    Log,
    LogShiftValue,
    Sqrt,
    SqrtShiftValue,
)

# We don't test
# - LogShiftTwo
# - SqrtShiftTwo
# at the moment since they derive from the ShiftValue

REAL_LINE_LINKS = [Identity]
POSITIVE_LINE_LINKS = [Log, Sqrt, LogIdent, InverseSoftPlus]
POSITIVE_RESTRICTED_LINE_LINKS = [Logit]  ##domain (0,1)
SHIFTED_LINKS = [LogShiftValue, SqrtShiftValue, InverseSoftPlusShiftValue]
VALUES = np.array([2, 5, 10, 25, 100])
M = 10000


@pytest.mark.parametrize("linkfun", REAL_LINE_LINKS)
def test_link_real_line(linkfun):
    """Test links that are defined on the real line (-inf, inf)"""
    instance = linkfun()
    x = np.linspace(-100, 100, M)
    y = instance.inverse(instance.link(x))
    print(y)
    assert np.allclose(x, y), "Links don't match"


@pytest.mark.parametrize("linkfun", [Logit])
def test_link_zero_one_domain(linkfun):
    """Test links that are defined on the (0, 1) domain"""
    instance = linkfun()
    # Avoid exact 0 and 1 to stay within the domain
    x = np.linspace(1e-10, 1 - 1e-10, M)
    y = instance.inverse(instance.link(x))
    assert np.allclose(x, y), "Links don't match"


@pytest.mark.parametrize("linkfun", POSITIVE_LINE_LINKS)
def test_link_positive_line(linkfun):
    """Test links that are defined on the positive real line (eps, inf)"""
    instance = linkfun()
    x = np.linspace(1e-10, 100, M)
    y = instance.inverse(instance.link(x))
    assert np.allclose(x, y), "Links don't match"


@pytest.mark.parametrize("linkfun", POSITIVE_RESTRICTED_LINE_LINKS)
def test_link_positive_restricted_line(linkfun):
    """Test links that are defined on (eps, 1-eps)"""
    instance = linkfun()
    eps = 1e-10
    x = np.linspace(eps, 1 - eps, M)
    y = instance.inverse(instance.link(x))
    assert np.allclose(x, y), "Links don't match"


@pytest.mark.parametrize("linkfun", SHIFTED_LINKS)
@pytest.mark.parametrize("value", VALUES)
def test_link_positive_shifted_line(linkfun, value):
    """Test links that are shifted. This changes the domain of the links."""
    instance = linkfun(value)
    x = np.linspace(value, 100 + value, M)
    y = instance.inverse(instance.link(x))
    assert np.allclose(x, y), "Links don't match"



import sys
import warnings
from typing import Any, Dict, Union

import numpy as np


def handle_param_dict(
    self: Any,
    param: Union[Any, Dict],
    default: Union[Any, Dict],
    n_params: int,
    name: str,
) -> None:
    """This function takes the potentially incomplete parameter dict given by the user
    and the default value for a GAMLSS param and will fill it for all missing distribution
    parameters with the default. If the user passes just a bool, string, int, it will
    create a matching dictionary.

    This is necessary since ondil expects at many places as many user given parameters
    as distribution parameters, but we don't know the number of distribution
    parameters a-priori at the estimator initialization.

    This function proceeds to set the default value!

    Args:
        self (Any): The self
        param (Union[Any, Dict]): The param given by the user
        default (Any): The default value
        n_params (int): The number of distribution parameters.
        name (str): The name of the parameter.

    Warns:
        Raises a warning if the user sets an incomplete default dict.

    """
    # Handle (incomplete) dictionaries of parameters.
    if isinstance(param, dict):
        for p in range(n_params):
            if p not in param.keys():
                warnings.warn(
                    f"[{self.__class__.__name__}] "
                    f"No value given for parameter {name} for distribution "
                    f"parameter {p}. Setting default value {default}.",
                    RuntimeWarning,
                    stacklevel=1,
                )
                if isinstance(default, dict):
                    param[p] = default[p]
                else:
                    param[p] = default
    else:
        # No warning since we expect that floats/strings/ints are either the defaults
        # Or given on purpose for all params the ame
        param = {p: param for p in range(n_params)}

    setattr(self, name, param)


def calculate_asymptotic_training_length(forget: float):
    if forget == 0:
        # Return the highest possible value that is no infinity
        return sys.maxsize
    else:
        return 1 / forget


def calculate_effective_training_length(forget: float, n_obs: int):
    if forget == 0:
        return n_obs
    else:
        return (1 - (1 - forget) ** n_obs) / forget


def _online_mean_update(avg: float, value: float, forget: float, n_seen: int):
    n_asymmptotic = calculate_asymptotic_training_length(forget)
    n_eff = calculate_effective_training_length(forget, n_seen)
    forget_scaled = forget * np.maximum(n_asymmptotic / n_eff, 1.0)
    diff = value - avg
    incr = forget_scaled * diff
    if forget_scaled > 0:
        new_avg = avg + incr
    else:
        new_avg = avg + diff / n_seen
    return new_avg


def online_mean_update(
    avg: float,
    value: float | np.ndarray,
    forget: float,
    n_seen: int,
) -> float:
    """Update the average with a new value or an array of values using an online mean update.

    Args:
        avg (float): The current average.
        value (float | np.ndarray): The new value or array of values to update the average with.
            If `value` is an array, the average is updated iteratively for each element.
        forget (float): The forgetting factor.
        n_seen (int): The number of observations seen so far.

    Returns:
        float: The updated average.
    """
    if isinstance(value, np.ndarray):
        _avg = float(avg)
        for i, v in zip(range(-value.shape[0] + 1, 1), value, strict=True):
            # Substract since we expect n_seen to be number of observations INCLUDING
            # the ones in values
            _avg = _online_mean_update(_avg, v, forget, n_seen + i)
        return _avg
    else:
        return _online_mean_update(avg, value, forget, n_seen)



from typing import Tuple

import numpy as np
import scipy.special as sp
import scipy.stats as st

from ..base import Distribution, LinkFunction, ScipyMixin
from ..links import Identity, Log, LogShiftTwo


class StudentT(ScipyMixin, Distribution):
    """Corresponds to GAMLSS TF() and scipy.stats.t()"""

    corresponding_gamlss: str = "TF"

    parameter_names = {0: "mu", 1: "sigma", 2: "nu"}
    parameter_support = {
        0: (-np.inf, np.inf),
        1: (np.nextafter(0, 1), np.inf),
        2: (np.nextafter(0, 1), np.inf),
    }
    distribution_support = (-np.inf, np.inf)

    # Scipy distribution and parameter mapping ondil -> scipy
    scipy_dist = st.t
    scipy_names = {"mu": "loc", "sigma": "scale", "nu": "df"}

    def __init__(
        self,
        loc_link: LinkFunction = Identity(),
        scale_link: LinkFunction = Log(),
        tail_link: LinkFunction = LogShiftTwo(),
    ) -> None:
        super().__init__(
            links={
                0: loc_link,
                1: scale_link,
                2: tail_link,
            }
        )

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma, nu = self.theta_to_params(theta)

        if param == 0:
            # MU
            s2 = sigma**2
            dsq = (y - mu) ** 2 / s2
            omega = (nu + 1) / (nu + dsq)
            return (omega * (y - mu)) / s2

        if param == 1:
            # SIGMA
            dsq = (y - mu) ** 2 / sigma**2
            omega = (nu + 1) / (nu + dsq)
            return (omega * dsq - 1) / sigma

        if param == 2:
            # TAIL
            dsq = (y - mu) ** 2 / sigma**2
            omega = (nu + 1) / (nu + dsq)
            dsq3 = 1 + (dsq / nu)
            v2 = nu / 2
            v3 = (nu + 1) / 2
            return (
                -np.log(dsq3)
                + ((omega * dsq - 1) / nu)
                + sp.digamma(v3)
                - sp.digamma(v2)
            ) / 2

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        _, sigma, nu = self.theta_to_params(theta)
        if param == 0:
            # MU
            return -(nu + 1) / ((nu + 3) * sigma**2)

        if param == 1:
            # SIGMA
            return -(2 * nu) / ((nu + 3) * sigma**2)

        if param == 2:
            # TAIL
            nu = np.fmin(nu, 1e15)
            v2 = nu / 2
            v3 = (nu + 1) / 2
            out = (  ## Polygamma(1, x) is the same as trigamma(x) in R
                (sp.polygamma(1, v3) - sp.polygamma(1, v2))
                + (2 * (nu + 5)) / (nu * (nu + 1) * (nu + 3))
            ) / 4
            return np.clip(out, -np.inf, -1e-15)

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        if sorted(params) == [0, 1]:
            # d2l/(dm ds)
            return np.zeros_like(y)

        if sorted(params) == [0, 2]:
            # d2l/(dm dn)
            return np.zeros_like(y)

        if sorted(params) == [1, 2]:
            # d2l / (dm dn)
            _, sigma, nu = self.theta_to_params(theta)
            return 2 / (sigma * (nu + 3) * (nu + 1))

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        params = st.t.fit(y)
        return np.tile([params[1], params[2], params[0]], (y.shape[0], 1))



from typing import Literal, Union
import numpy as np


class InformationCriterion:
    """Calculate the information criteria.

    +---------+--------------------------------------+--------------------------+
    | `ic`    | Information Criterion                | Formula                  |
    +=========+======================================+==========================+
    | `"aic"` | Akaike's Information Criterion       | $- 2l + 2p$              |
    | `"aicc"`| Corr. Akaike's Information Criterion | $- 2l + 2pn/(n-p-1)$     |
    | `"bic"` | Bayesian Information Criterion       | $- 2l + p\log(n)$        |
    | `"hqc"` | Hannan-Quinn Information Criterion   | $- 2l + 2p\log(\log(n))$ |
    | `"max"` | Select the largest model             |                          |
    +---------+--------------------------------------+--------------------------+

    Methods:
    -------
    from_rss(rss)
        Compute the chosen criterion from residual sum of squares.
    from_ll(log_likelihood)
        Compute the chosen criterion directly from a log-likelihood value.
    """

    def __init__(
        self,
        n_observations: Union[int, np.ndarray],
        n_parameters: Union[int, np.ndarray],
        criterion: Literal["aic", "bic", "hqc", "aicc", "max"] = "aic",
    ):
        """
        Args:
            n_observations (int or array-like): Number of observations used in the model.
            n_parameters (int or array-like): Number of estimated parameters in the model.
            criterion ({"aic","bic","hqc","aicc", "max"}, default="aic"): The information criterion to compute.


        Raises:
            ValueError: If the criterion is not recognized.
        """
        # validate criterion early
        if criterion not in ("aic", "aicc", "bic", "hqc", "max"):
            raise ValueError(f"Criterion '{criterion}' not recognized.")
        self.n_observations = n_observations
        self.n_parameters = n_parameters
        self.criterion = criterion

    def from_rss(self, rss: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """
        Compute the specified information criterion from the residual sum of squares (RSS).

        The Gaussian log-likelihood is estimated as:
            ll = -n/2 * log(rss / n) - n/2 * (1 + log(2π))

        Args:
            rss (float or array-like): Residual sum of squares of the fitted model.

        Returns:
            ic (float or array-like): The information criterion value (AIC, AICC, BIC, HQC, or Max).
        """
        # Gaussian log‐likelihood: ll = -n/2*log(rss/n) + constant
        # https://en.wikipedia.org/wiki/Akaike_information_criterion#Comparison_with_least_squares
        constant_term = -self.n_observations / 2 * (1 + np.log(2 * np.pi))
        log_likelihood = (
            -self.n_observations / 2 * np.log(rss / self.n_observations) + constant_term
        )
        return self.from_ll(log_likelihood)

    def from_ll(
        self, log_likelihood: Union[float, np.ndarray]
    ) -> Union[float, np.ndarray]:
        """
        Compute the specified information criterion directly from log-likelihood.

        Args:
            log_likelihood (float or array-like): The log-likelihood of the model.

        Returns:
            ic (float or array-like): The information criterion value (AIC, AICC, BIC, HQC, or Max).
        """
        c = self.criterion
        n, p, ll = self.n_observations, self.n_parameters, log_likelihood
        if c == "aic":
            return - 2 * ll + p * 2 
        elif c == "aicc":
            if n - p - 1 == 0:
                raise ValueError("Invalid inputs: n - p - 1 must not be zero for AICC calculation.")
            return - 2 * ll + p * 2 * n / ( n - p - 1)
        elif c == "bic":
            return - 2 * ll + p * np.log(n) 
        elif c == "hqc":
            return - 2 * ll + p * 2 * np.log(np.log(n))
        elif c == "max":
            return -ll



from typing import Optional

import numpy as np

from ..base import EstimationMethod
from ..gram import init_inverted_gram, init_y_gram, update_inverted_gram, update_y_gram


class OrdinaryLeastSquares(EstimationMethod):
    """Simple ordinary least squares respectively recursive least squares. No fancy parameters possible."""

    def __init__(self) -> None:
        super().__init__(
            _path_based_method=False,
            _accepts_bounds=False,
            _accepts_selection=False,
        )

    @staticmethod
    def init_x_gram(X: np.ndarray, weights: np.ndarray, forget: float) -> np.ndarray:
        return init_inverted_gram(X, w=weights, forget=forget)

    @staticmethod
    def init_y_gram(
        X: np.ndarray, y: np.ndarray, weights: np.ndarray, forget: float
    ) -> np.ndarray:
        return init_y_gram(X, y, w=weights, forget=forget)

    @staticmethod
    def update_x_gram(
        gram: np.ndarray, X: np.ndarray, weights: np.ndarray, forget: float
    ) -> np.ndarray:
        return update_inverted_gram(gram, X, w=weights, forget=forget)

    @staticmethod
    def update_y_gram(
        gram: np.ndarray,
        X: np.ndarray,
        y: np.ndarray,
        weights: np.ndarray,
        forget: float,
    ) -> np.ndarray:
        return update_y_gram(gram, X, y, forget=forget, w=weights)

    def fit_beta_path(
        self, x_gram: np.ndarray, y_gram: np.ndarray, is_regularized: np.ndarray
    ) -> np.ndarray:
        return super().fit_beta_path(x_gram, y_gram, is_regularized)

    def update_beta_path(
        self,
        x_gram: np.ndarray,
        y_gram: np.ndarray,
        beta_path: np.ndarray,
        is_regularized: np.ndarray,
    ) -> np.ndarray:
        return super().update_beta_path(x_gram, y_gram, beta_path, is_regularized)

    def fit_beta(
        self,
        x_gram: np.ndarray,
        y_gram: np.ndarray,
        is_regularized: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        return (x_gram @ y_gram).squeeze(-1)

    def update_beta(
        self,
        x_gram: np.ndarray,
        y_gram: np.ndarray,
        beta: np.ndarray,
        is_regularized: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        return (x_gram @ y_gram).squeeze(-1)



from typing import Optional, Tuple
import numpy as np
import scipy.stats as st

from ..base import Distribution, LinkFunction, ScipyMixin
from ..links import Identity, Log


class ReverseGumbel(ScipyMixin, Distribution):
    """
    The Reverse Gumbel (Type I minimum extreme value) distribution with location (mu) and scale (sigma) parameters.

    The probability density function is defined as:
    $$
    f(y | \\mu, \\sigma) = \\frac{1}{\\sigma} \\exp\\left( \\frac{y - \\mu}{\\sigma} - \\exp\\left( \\frac{y - \\mu}{\\sigma} \\right) \\right)
    $$

    This distribution corresponds to the RG() distribution in GAMLSS.

    Notes:
        - Mean = mu - digamma(1) * sigma ≈ mu - 0.5772157 * sigma
        - Variance = (pi^2 * sigma^2) / 6 ≈ 1.64493 * sigma^2
    """

    corresponding_gamlss: str = "RG"
    parameter_names = {0: "mu", 1: "sigma"}
    parameter_support = {0: (-np.inf, np.inf), 1: (np.nextafter(0, 1), np.inf)}
    distribution_support = (-np.inf, np.inf)
    scipy_dist = st.gumbel_r
    scipy_names = {"mu": "loc", "sigma": "scale"}

    def __init__(
        self,
        loc_link: LinkFunction = Identity(),
        scale_link: LinkFunction = Log(),
    ) -> None:
        super().__init__(
            links={
                0: loc_link,
                1: scale_link,
            }
        )

    def theta_to_scipy_params(self, theta: np.ndarray) -> dict:
        mu = theta[:, 0]
        sigma = theta[:, 1]
        return {
            "loc": mu,
            "scale": sigma,
        }

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma = self.theta_to_params(theta)
        z = (y - mu) / sigma
        ez = np.exp(-z)

        if param == 0:
            return (1 - ez) / sigma

        if param == 1:
            return -(1 / sigma) * (1 - z) - (z / sigma) * ez

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        _, sigma = self.theta_to_params(theta)

        if param == 0:
            return -1 / sigma**2

        if param == 1:
            return -1.82368 / sigma**2  # approximation from GAMLSS

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        _, sigma = self.theta_to_params(theta)
        return np.full_like(y, -0.422784 / sigma**2)

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        gamma_const = 0.5772157
        sigma_init = (np.sqrt(6) * np.std(y)) / np.pi
        mu_init = np.mean(y) + gamma_const * sigma_init
        initial_params = [mu_init, sigma_init]
        return np.tile(initial_params, (y.shape[0], 1))



from .distribution import Distribution, ScipyMixin
from .estimation_method import EstimationMethod
from .estimator import Estimator, OndilEstimatorMixin
from .link import LinkFunction

__all__ = [
    "Distribution",
    "ScipyMixin",
    "LinkFunction",
    "Estimator",
    "EstimationMethod",
    "OndilEstimatorMixin",
]



from typing import Tuple

import numpy as np

from ..base import LinkFunction
from .robust_math import SMALL_NUMBER


class Sqrt(LinkFunction):
    """The square root Link function.

    The square root link function is defined as $$g(x) = \sqrt(x)$$.
    """

    link_support = (np.nextafter(0, 1), np.inf)

    def __init__(self):
        pass

    def link(self, x: np.ndarray) -> np.ndarray:
        return np.sqrt(x)

    def inverse(self, x: np.ndarray) -> np.ndarray:
        return np.power(x, 2)

    def inverse_derivative(self, x: np.ndarray) -> np.ndarray:
        return 2 * x

    def link_derivative(self, x: np.ndarray) -> np.ndarray:
        return 1 / (2 * np.sqrt(x))

    def link_second_derivative(self, x) -> np.ndarray:
        return -1 / (4 * x ** (3 / 2))


class SqrtShiftValue(LinkFunction):
    """
    The Sqrt-Link function shifted to a value \(v\).

    This link function is defined as $$g(x) = \sqrt(x - v)$$. It can be used
    to ensure that certain distribution paramters don't fall below lower
    bounds, e.g. ensuring that the degrees of freedom of a Student's t distribtuion
    don't fall below 2, hence ensuring that the variance exists.
    """

    def __init__(self, value: float):
        self.value = value

    @property
    def link_support(self) -> Tuple[float, float]:
        return (self.value + np.nextafter(0, 1), np.inf)

    def link(self, x: np.ndarray) -> np.ndarray:
        return np.sqrt(x - self.value + SMALL_NUMBER)

    def inverse(self, x: np.ndarray) -> np.ndarray:
        return self.value + np.power(x, 2)

    def inverse_derivative(self, x: np.ndarray) -> np.ndarray:
        return 2 * x

    def link_derivative(self, x: np.ndarray) -> np.ndarray:
        return 1 / (2 * np.sqrt(x - self.value + SMALL_NUMBER))

    def link_second_derivative(self, x) -> np.ndarray:
        return -1 / (4 * (x - self.value + SMALL_NUMBER) ** (3 / 2))


class SqrtShiftTwo(SqrtShiftValue):
    """
    The Sqrt-Link function shifted to 2.

    This link function is defined as $$g(x) = \sqrt(x - 2)$$. It can be used
    to ensure that certain distribution paramters don't fall below lower
    bounds, e.g. ensuring that the degrees of freedom of a Student's t distribtuion
    don't fall below 2, hence ensuring that the variance exists.
    """

    def __init__(self):
        super().__init__(2)
        pass



import warnings
from abc import ABC, abstractmethod
from typing import Dict, Optional, Tuple

import numpy as np
import scipy.optimize as opt
import scipy.stats as st

from ..warnings import OutOfSupportWarning
from .link import LinkFunction


class Distribution(ABC):
    def __init__(self, links: dict[int, LinkFunction]) -> None:
        self.links = links
        self._validate_links()

    def __call__(self, *args, **kwds):
        raise NotImplementedError("Not implemented but necessary for sklearn.")

    @property
    def corresponding_gamlss(self) -> str | None:
        """The name of the corresponding implementation in 'gamlss.dist' R package."""
        pass

    @property
    @abstractmethod
    def parameter_names(self) -> dict:
        """Parameter name for each column of theta."""
        pass

    @property
    def n_params(self) -> int:
        """Each subclass must define 'n_params'."""
        return len(self.parameter_names)

    def theta_to_params(self, theta: np.ndarray) -> Tuple[np.ndarray, ...]:
        """Take the fitted values and return tuple of vectors for distribution parameters."""
        return tuple(theta[:, i] for i in range(self.n_params))

    @property
    @abstractmethod
    def distribution_support(self) -> Tuple[float, float]:
        """The support of the distribution."""
        pass

    @property
    @abstractmethod
    def parameter_support(self) -> dict:
        """The support of each parameter of the distribution."""
        pass

    @abstractmethod
    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int) -> np.ndarray:
        """Take the first derivative of the likelihood function with respect to the param."""

    @abstractmethod
    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int) -> np.ndarray:
        """Take the second derivative of the likelihood function with respect to the param."""

    @abstractmethod
    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int]
    ) -> np.ndarray:
        """Take the first derivative of the likelihood function with respect to both parameters."""

    def _validate_links(self):
        for param, link in self.links.items():
            if link.link_support[0] < self.parameter_support[param][0]:
                warnings.warn(
                    message=f"Lower bound of link function is smaller than the parameter support for parameter {param} ",
                    category=OutOfSupportWarning,
                )
            if link.link_support[1] > self.parameter_support[param][1]:
                warnings.warn(
                    message=f"Upper bound of link function is larger than the parameter support for parameter {param} ",
                    category=OutOfSupportWarning,
                )

    def _validate_dln_dpn_inputs(
        self, y: np.ndarray, theta: np.ndarray, param: int
    ) -> None:
        if param >= self.n_params:
            raise ValueError(
                f"{self.__class__.__name__} has only {self.n_params} distribution parameters.\nYou have passed {param}. Please remember we start counting at 0."
            )

    def _validate_dl2_dpp_inputs(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int]
    ) -> None:
        if max(params) >= self.n_params:
            raise ValueError(
                f"{self.__class__.__name__} has only {self.n_params} distribution parameters.\nYou have passed {params}. Please remember we start counting at 0."
            )
        if params[0] == params[1]:
            raise ValueError("Cross derivatives must use different parameters.")

    def link_function(self, y: np.ndarray, param: int = 0) -> np.ndarray:
        """Apply the link function for param on y."""
        return self.links[param].link(y)

    def link_inverse(self, y: np.ndarray, param: int = 0) -> np.ndarray:
        """Apply the inverse of the link function for param on y."""
        return self.links[param].inverse(y)

    def link_function_derivative(self, y: np.ndarray, param: int = 0) -> np.ndarray:
        """Apply the derivative of the link function for param on y."""
        return self.links[param].link_derivative(y)

    def link_inverse_derivative(self, y: np.ndarray, param: int = 0) -> np.ndarray:
        """Apply the derivative of the inverse link function for param on y."""
        return self.links[param].inverse_derivative(y)

    @abstractmethod
    def initial_values(
        self, y: np.ndarray, param: int = 0, axis: Optional[int | None] = None
    ) -> np.ndarray:
        """Calculate the initial values for the GAMLSS fit."""

    def quantile(self, q: np.ndarray, theta: np.ndarray) -> np.ndarray:
        """
        Compute the quantile function for the given data.

        This is a alias for the `ppf` method.

        Args:
            q (np.ndarray): The quantiles to compute.
            theta (np.ndarray): The parameters of the distribution.

        Returns:
            np.ndarray: The quantiles corresponding to the given probabilities.
        """
        return self.ppf(q, theta)

    @abstractmethod
    def calculate_conditional_initial_values(
        self,
        y: np.ndarray,
        theta: np.ndarray,
        param: int,
    ) -> np.ndarray:
        """Calculate the conditional initial values for the GAMLSS fit."""

    @abstractmethod
    def cdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        """
        Compute the cumulative distribution function (CDF) for the given data.

        Args:
            y (np.ndarray): The data points at which to evaluate the CDF.
            theta (np.ndarray): The parameters of the distribution.

        Returns:
            np.ndarray: The CDF evaluated at the given data points.
        """

    @abstractmethod
    def pdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        """
        Compute the probability density function (PDF) for the given data points.

        Args:
            y (np.ndarray): An array of data points at which to evaluate the PDF.
            theta (np.ndarray): An array of parameters for the distribution.

        Returns:
            np.ndarray: An array of PDF values corresponding to the data points in `y`.
        """

    @abstractmethod
    def pmf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        """
        Compute the probability mass function (PMF) for the given data points.

        Args:
            y (np.ndarray): An array of data points at which to evaluate the PDF.
            theta (np.ndarray): An array of parameters for the distribution.

        Returns:
            np.ndarray: An array of PMF values corresponding to the data points in `y`.
        """

    @abstractmethod
    def ppf(self, q: np.ndarray, theta: np.ndarray) -> np.ndarray:
        """
        Percent Point Function (Inverse of CDF).

        Args:
            q (np.ndarray): Quantiles.
            theta (np.ndarray): Distribution parameters.

        Returns:
            np.ndarray: The quantile corresponding to the given probabilities.
        """

    @abstractmethod
    def rvs(self, size: int, theta: np.ndarray) -> np.ndarray:
        """
        Generate random variates of given size and parameters.

        Args:
            size (int): The number of random variates to generate.
            theta (np.ndarray): The parameters for the distribution.

        Returns:
            np.ndarray: A 2D array of random variates with shape (theta.shape[0], size).
        """

    @abstractmethod
    def logpmf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        raise NotImplementedError(
            "Log PMF is not implemented for continuous distributions."
        )

    @abstractmethod
    def logpdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        raise NotImplementedError(
            "Log PDF is not implemented for discrete distributions."
        )

    @abstractmethod
    def logcdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        """Compute the log of the cumulative distribution function (CDF) for the given data points.

        Args:
            y (np.ndarray): An array of data points at which to evaluate the log CDF.
            theta (np.ndarray): An array of parameters for the distribution.

        Returns:
            np.ndarray: An array of log CDF values corresponding to the data points in `y`.
        """


class ScipyMixin(ABC):
    @property
    @abstractmethod
    def parameter_names(self) -> dict:
        """Parameter name for each column of theta."""
        pass

    @property
    @abstractmethod
    def scipy_dist(self) -> st.rv_continuous:
        """The names of the parameters in the scipy.stats distribution and the corresponding column in theta."""
        pass

    @property
    @abstractmethod
    def scipy_names(self) -> Tuple[str]:
        """The names of the parameters in the scipy.stats distribution and the corresponding column in theta."""
        pass

    def theta_to_scipy_params(self, theta: np.ndarray) -> Dict[str, np.ndarray]:
        """Maps $\\theta$ to the `scipy` parameters.

        Args:
            theta (np.ndarray): $\\theta$ as estimated by `OnlineDistributionalRegression()` estimator

        Raises:
            ValueError: If we don't define the `scipy_names` attribute.

        Returns:
            dict: Dictionary that can be unrolled into scipy distribution class as in `st.some_dist(**return_value)`
        """
        if not self.scipy_names:
            raise ValueError(
                f"{self.__class__.__name__} has no scipy_names defined. To use theta_to_scipy_params Please define them in the subclass. Or override this method in the subclass if there is no 1:1 mapping between theta columns and scipy params."
            )

        params = {}
        for idx, name in self.parameter_names.items():
            params[self.scipy_names[name]] = theta[:, idx]
        return params

    def cdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        return self.scipy_dist(**self.theta_to_scipy_params(theta)).cdf(y)

    def pdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        return self.scipy_dist(**self.theta_to_scipy_params(theta)).pdf(y)

    def pmf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        raise NotImplementedError(
            "PMF is not implemented for continuous distributions."
        )

    def ppf(self, q: np.ndarray, theta: np.ndarray) -> np.ndarray:
        return self.scipy_dist(**self.theta_to_scipy_params(theta)).ppf(q)

    def logpmf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        """Compute the log of the probability mass function (PMF) for the given data points.

        Args:
            y (np.ndarray): An array of data points at which to evaluate the log PMF.
            theta (np.ndarray): An array of parameters for the distribution.

        Returns:
            np.ndarray: An array of log PMF values corresponding to the data points in `y`.
        """
        return self.scipy_dist(**self.theta_to_scipy_params(theta)).logpmf(y)

    def logpdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        """Compute the log of the probability density function (PDF) for the given data points.

        Args:
            y (np.ndarray): An array of data points at which to evaluate the log PDF.
            theta (np.ndarray): An array of parameters for the distribution.

        Returns:
            np.ndarray: An array of log PDF values corresponding to the data points in `y`.
        """
        return self.scipy_dist(**self.theta_to_scipy_params(theta)).logpdf(y)

    def logcdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        """Compute the log of the cumulative distribution function (CDF) for the given data points.

        Args:
            y (np.ndarray): An array of data points at which to evaluate the log CDF.
            theta (np.ndarray): An array of parameters for the distribution.

        Returns:
            np.ndarray: An array of log CDF values corresponding to the data points in `y`.
        """
        return self.scipy_dist(**self.theta_to_scipy_params(theta)).logcdf(y)

    def rvs(self, size: int, theta: np.ndarray) -> np.ndarray:
        return (
            self.scipy_dist(**self.theta_to_scipy_params(theta))
            .rvs((size, theta.shape[0]))
            .T
        )

    def mean(self, theta: np.ndarray) -> np.ndarray:
        """Compute the mean of the distribution for the given parameters.

        Args:
            theta (np.ndarray): An array of parameters for the distribution.

        Returns:
            np.ndarray: An array of means corresponding to the parameters in `theta`.
        """
        return self.scipy_dist(**self.theta_to_scipy_params(theta)).mean()

    def _scipy_mle_objective(
        self,
        start: np.ndarray,
        y: np.ndarray,
        fixed: dict,
        param: int,
    ):
        unknown = np.arange(param, self.n_params)
        decision = {}
        for i, value in zip(unknown, start):
            decision[self.scipy_names[self.parameter_names[i]]] = value
        return -np.sum(self.scipy_dist(**decision, **fixed).logpdf(y))

    def calculate_conditional_initial_values(
        self,
        y: np.ndarray,
        theta: np.ndarray,
        param: int,
    ):
        start = theta[0, param:]
        fixed = self.theta_to_scipy_params(theta)
        unknown = np.arange(param, self.n_params)
        for i in unknown:
            del fixed[self.scipy_names[self.parameter_names[i]]]

        bounds = [self.links[b].link_support for b in unknown]
        result = opt.minimize(
            self._scipy_mle_objective,
            x0=start,
            args=(y, fixed, param),
            method="L-BFGS-B",
            bounds=bounds,
            # options={"disp": True}
        )
        new = theta.copy()
        if not result.success:
            print("Optimization failed: " + result.message)
            print("Will use unconditional start values")
        else:
            new[:, unknown] = np.expand_dims(result.x, 0)
        return new



import numpy as np
import pytest
from sklearn.datasets import make_regression

from ondil.distributions import JSU
from ondil.estimators import OnlineDistributionalRegression

FIT_INTERCEPT = [True, False]
N_FEATURES = np.round(np.geomspace(11, 100, 5)).astype(int)


@pytest.mark.parametrize("n_features", N_FEATURES, ids=lambda x: f"n_features_{x}")
@pytest.mark.parametrize(
    "fit_intercept", FIT_INTERCEPT, ids=lambda x: f"fit_intercept_{x}"
)
def test_get_J_from_equation(n_features, fit_intercept):
    equation = {
        0: "all",  # should adjust to n_features
        1: "intercept",
        2: np.arange(0, 4),
        3: np.array([True] * 10 + [False] * (n_features - 10)).astype(bool),
    }

    EXPECTED = {
        0: {True: n_features + 1, False: n_features},
        1: {True: 1, False: 1},
        2: {True: 5, False: 4},
        3: {True: 11, False: 10},
    }

    X, _ = make_regression(n_samples=100, n_features=n_features)
    distribution = JSU()

    estimator = OnlineDistributionalRegression(
        distribution=distribution,
        equation=equation,
        fit_intercept=fit_intercept,
    )
    estimator._prepare_estimator()
    J = estimator.get_J_from_equation(X)
    for param, expected_dict in EXPECTED.items():
        assert J[param] == expected_dict[fit_intercept], f"Wrong J for param == {param}"


def test_get_J_from_equation_warnings():
    n_features = 10
    fit_intercept = True

    equation_fail_2 = {
        0: "all",  # should adjust to n_features
        1: "intercept",
        2: np.arange(0, 20),
        3: np.array([True] * n_features).astype(bool),
    }

    X, _ = make_regression(n_samples=100, n_features=n_features)
    distribution = JSU()

    estimator = OnlineDistributionalRegression(
        distribution=distribution,
        equation=equation_fail_2,
        fit_intercept=fit_intercept,
    )
    estimator._prepare_estimator()
    with pytest.raises(ValueError, match="Shape does not match for param 2."):
        _ = estimator.get_J_from_equation(X)

    # Test for parameter three
    equation_fail_3 = {
        0: "all",  # should adjust to n_features
        1: "intercept",
        2: np.arange(0, n_features),
        3: np.array([True, False] * 10).astype(bool),
    }

    X, _ = make_regression(n_samples=100, n_features=10)
    distribution = JSU()
    estimator = OnlineDistributionalRegression(
        distribution=distribution,
        equation=equation_fail_3,
        fit_intercept=fit_intercept,
    )
    estimator._prepare_estimator()

    with pytest.raises(
        ValueError,
        match="Shape does not match for param 3.",
    ):
        _ = estimator.get_J_from_equation(X)



from typing import Tuple
import numpy as np
import scipy.stats as st

from ..base import Distribution, LinkFunction, ScipyMixin
from ..links import Log


class Exponential(ScipyMixin, Distribution):
    """
    The Exponential distribution parameterized by the mean (mu).

    PDF:
    f(y | mu) = (1 / mu) * exp(-y / mu), for y > 0, mu > 0

    This corresponds to EXP() in GAMLSS where:
    - mu > 0
    - y > 0

    """

    corresponding_gamlss: str = "EXP"
    parameter_names = {0: "mu"}
    parameter_support = {0: (np.nextafter(0, 1), np.inf)}
    distribution_support = (np.nextafter(0, 1), np.inf)
    scipy_dist = st.expon
    scipy_names = {"mu": "scale"}

    def __init__(self, mu_link: LinkFunction = Log()) -> None:
        assert isinstance(mu_link, LinkFunction), "mu_link must be a LinkFunction"
        super().__init__(links={0: mu_link})

    def theta_to_scipy_params(self, theta: np.ndarray) -> dict:
        (mu,) = self.theta_to_params(theta)
        return {"scale": mu}

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        (mu,) = self.theta_to_params(theta)
        return (y - mu) / mu**2

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        (mu,) = self.theta_to_params(theta)
        return -1 / mu**2

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 0)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        return np.zeros_like(y)

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        return np.full((y.shape[0], 1), np.mean(y))



from typing import Tuple

import numpy as np

from ..base import LinkFunction
from .robust_math import (
    robust_exp,
    robust_softplus,
    robust_softplus_inverse,
    zero_safe_division,
)


class InverseSoftPlus(LinkFunction):
    """
    The softplus is defined as $$
        \operatorname{SoftPlus(x)} = \log(1 + \exp(x))
    $$ and hence the inverse is defined as $$
        \log(\exp(x) - 1)
    $$ which can be used as link function for the parameters on the
    positive real line. The behavior of the inverse softplus is more
    graceful on large values as it avoids the exp of the log-link and
    converges towards a linear behaviour.

    The softplus is the smooth approximation of $\max(x, 0)$.
    """

    link_support = (np.nextafter(0, 1), np.inf)

    def __init__(self):
        pass

    def link(self, x: np.ndarray):
        return robust_softplus_inverse(x)

    def inverse(self, x: np.ndarray):
        return robust_softplus(x)

    def link_derivative(self, x: np.ndarray):
        return zero_safe_division(1, (robust_exp(x) - 1) + 1)

    def link_second_derivative(self, x: np.ndarray):
        # return 1 / (2 - 2 * np.cosh(x)) # equivalent
        # return -1 / (np.exp(x) - 1) - 1 / (np.exp(x) - 1) ** 2
        return zero_safe_division(1, (2 - 2 * np.cosh(x)))

    def inverse_derivative(self, x: np.ndarray):
        return zero_safe_division(1, (robust_exp(-x) + 1))


class InverseSoftPlusShiftValue(LinkFunction):
    """
    The Inverse SoftPlus function shifted to a value \(v\).
    """

    def __init__(self, value: float):
        self.value = value

    @property
    def link_support(self) -> Tuple[float, float]:
        return (self.value + np.nextafter(0, 1), np.inf)

    def link(self, x: np.ndarray):
        z = x - self.value
        # return np.log1p(np.exp(-np.abs(z))) + np.maximum(z, 0)
        # return robust_log(robust_exp(z) - 1)
        return robust_softplus_inverse(z)

    def inverse(self, x: np.ndarray):
        return robust_softplus(x) + self.value

    def inverse_derivative(self, x: np.ndarray) -> np.ndarray:
        return zero_safe_division(1, (robust_exp(-x) + 1))

    def link_derivative(self, x: np.ndarray) -> np.ndarray:
        return zero_safe_division(1, (1 - robust_exp(self.value - x)))

    def link_second_derivative(self, x: np.ndarray) -> np.ndarray:
        return zero_safe_division(1, (2 - 2 * np.cosh(self.value - x)))


class InverseSoftPlusShiftTwo(InverseSoftPlusShiftValue):
    """
    The Inverse SoftPlus function shifted to 2.
    """

    def __init__(self):
        super().__init__(2)
        pass



import numpy as np
import pytest

from ondil.gram import init_forget_vector
from ondil.scaler import OnlineScaler


@pytest.mark.parametrize("N_init", [5, 10, 100], ids=["init_1", "init_10", "init_100"])
@pytest.mark.parametrize("D", [1, 10], ids=["vars_1", "vars_10"])
@pytest.mark.parametrize(
    "selection_dtype", [bool, int], ids=["sel_dtype_bool", "sel_dtype_int"]
)
@pytest.mark.parametrize(
    "forget",
    [0, 0.01, 0.1],
    ids=["no_forgetting", "forget_0.01", "forget_0.1"],
)
@pytest.mark.parametrize(
    "sample_weight",
    [True, False],
    ids=["sample_weight", "no_sample_weight"],
)
def test_online_scaler(N_init, D, forget, selection_dtype, sample_weight):
    """Test OnlineScaler with various combinations of parameters."""
    # Skip invalid combinations
    N = 100

    X = np.random.uniform(0, 10, N * D).reshape(-1, D)
    X_init = X[0:N_init,]

    # Calculate effective weights for comparison
    W_forget = init_forget_vector(forget, N).reshape(-1, 1)

    if sample_weight:
        W_sample = np.random.uniform(0.1, 1.0, (N, 1)).reshape(-1, 1)
        W_effective = W_forget * W_sample
    else:
        W_sample = None
        W_effective = W_forget

    if selection_dtype is bool:
        to_scale = np.random.choice([True, False], D)
    if selection_dtype is int:
        to_scale = np.random.choice(
            np.arange(D), np.random.randint(0, D + 1), replace=False
        )

    # Calculate true mean and variance for all columns after initial fit
    true_mean_init = np.array(
        [
            np.average(X_init[:, d], weights=W_effective[0:N_init].flatten())
            for d in range(D)
        ]
    )
    true_var_init = np.array(
        [
            np.average(
                (X_init[:, d] - true_mean_init[d]) ** 2,
                weights=W_effective[0:N_init].flatten(),
            )
            for d in range(D)
        ]
    )

    # Setup and fit the OnlineScaler
    os = OnlineScaler(forget=forget, to_scale=to_scale)
    if sample_weight:
        os.fit(X_init, sample_weight=W_sample[0:N_init,].flatten())
    else:
        os.fit(X_init)

    # # Assert initial fit is correct
    assert np.allclose(
        os.mean_, true_mean_init[to_scale]
    ), f"Initial mean mismatch: {os.mean_} vs {true_mean_init[to_scale]}"
    assert np.allclose(
        os.var_, true_var_init[to_scale]
    ), f"Initial variance mismatch: {os.var_} vs {true_var_init[to_scale]}"

    # For N_init = 1, var = 0, so scaling is not defined
    if N_init > 1:
        expected_out = np.copy(X_init)
        scaled_out = (X_init - true_mean_init) / np.sqrt(true_var_init)
        expected_out[:, to_scale] = scaled_out[:, to_scale]
        out = os.transform(X=X_init)
        assert np.allclose(
            out, expected_out
        ), f"Initial scaled X mismatch: {out} vs {expected_out}"

    # Test partial fit updates
    for i in range(N_init, N):
        if sample_weight:
            os.update(X[i : i + 1,], sample_weight=W_sample[i,])
        else:
            os.update(X[i : i + 1,])

        # Calculate true mean and variance for all columns
        true_mean = np.array(
            [
                np.average(
                    X[0 : (i + 1), d], weights=W_effective[0 : (i + 1)].flatten()
                )
                for d in range(D)
            ]
        )
        true_var = np.array(
            [
                np.average(
                    (X[0 : (i + 1), d] - true_mean[d]) ** 2,
                    weights=W_effective[0 : (i + 1)].flatten(),
                )
                for d in range(D)
            ]
        )
        expected_out = np.copy(X[0 : (i + 1),])
        scaled_out = (X[0 : (i + 1), :] - true_mean) / np.sqrt(true_var)
        expected_out[:, to_scale] = scaled_out[:, to_scale]
        out = os.transform(X=X[0 : (i + 1), :])

        assert np.allclose(
            os.mean_, true_mean[to_scale]
        ), f"Step {i - N_init + 1}: Mean mismatch: {os.mean_} vs {true_mean[to_scale]}"
        assert np.allclose(
            os.var_, true_var[to_scale]
        ), f"Step {i - N_init + 1}: Variance mismatch: {os.var_} vs {true_var[to_scale]}"
        assert np.allclose(
            out, expected_out
        ), f"Scaled X mismatch: {out} vs {expected_out}"


@pytest.mark.parametrize("D", [1, 10], ids=["features_1", "features_10"])
def test_standard_scaling_dont_scale(D):
    N = 100
    X = np.random.uniform(0, 10, N * D).reshape(-1, D)
    expected_out = X

    scaler = OnlineScaler(forget=0, to_scale=False)
    scaler.fit(X=X)
    out = scaler.transform(X=X)
    np.testing.assert_array_almost_equal(expected_out, out)



from abc import ABC, abstractmethod


class EstimationMethod(ABC):
    def __init__(
        self,
        _path_based_method,
        _accepts_bounds,
        _accepts_selection,
    ):
        self._path_based_method = _path_based_method
        self._accepts_bounds = _accepts_bounds
        self._accepts_selection = _accepts_selection

    @abstractmethod
    def init_x_gram(self, X, weights, forget):
        pass

    @abstractmethod
    def init_y_gram(self, X, y, weights, forget):
        pass

    @abstractmethod
    def update_x_gram(self, gram, X, weights, forget):
        pass

    @abstractmethod
    def update_y_gram(self, gram, X, y, weights, forget):
        pass

    @abstractmethod
    def fit_beta(self, x_gram, y_gram, is_regularized):
        if self._path_based_method:
            raise NotImplementedError("Method does not support non-path-based fitting.")

    @abstractmethod
    def update_beta(self, x_gram, y_gram, beta, is_regularized):
        if self._path_based_method:
            raise NotImplementedError("Method does not support non-path-based fitting.")

    @abstractmethod
    def fit_beta_path(self, x_gram, y_gram, is_regularized):
        if not self._path_based_method:
            raise NotImplementedError("Method does not support path-based fitting.")

    @abstractmethod
    def update_beta_path(self, x_gram, y_gram, beta_path, is_regularized):
        if not self._path_based_method:
            raise NotImplementedError("Method does not support path-based fitting.")



from typing import Tuple

import numpy as np
import scipy.stats as st

from ..base import Distribution, LinkFunction, ScipyMixin
from ..links import Identity, Log
from ..links.robust_math import robust_exp


class LogNormal(ScipyMixin, Distribution):
    """
    The Log-Normal distribution with mean and standard deviation parameterization in the log-space.

    The probability density function of the distribution is defined as:
    $$
    f(y | \\mu, \\sigma) = \\frac{1}{y\\sigma\\sqrt{2\\pi}}\\exp\\left(-\\frac{(\log y - \\mu)^2}{2\\sigma^2}\\right).
    $$
     respectively
    $$
    f(y | \\theta_0, \\theta_1) = \\frac{1}{y\\theta_1\\sqrt{2\pi}}\exp\\left(-\\frac{(\\log y - \\theta_0)^2}{2\\theta_1^2}\\right).
    $$
    where $y$ is the observed data, $\\mu = \\theta_0$ is the location parameter and $\\sigma = \\theta_1$ is the scale parameter.

    !!! note
        Note that re-parameterization used to move from scipy.stats to GAMLSS is:
        $$
            \\mu = \\exp(\\theta_0)
        $$
        and can therefore be numerically unstable for large values of $\\theta_0$.
        We have re-implemented the PDF, CDF, PPF according to avoid this issue,
        however the rvs method still uses the scipy.stats implementation which is not
        numerically stable for large values of $\\theta_0$.

    """

    corresponding_gamlss: str = "LOGNO"
    parameter_names = {0: "mu", 1: "sigma"}
    parameter_support = {0: (-np.inf, np.inf), 1: (np.nextafter(0, 1), np.inf)}
    distribution_support = (0, np.inf)
    scipy_dist = st.lognorm
    scipy_names = {"mu": "scale", "sigma": "s"}

    def __init__(
        self,
        loc_link: LinkFunction = Identity(),
        scale_link: LinkFunction = Log(),
    ) -> None:
        super().__init__(
            links={
                0: loc_link,
                1: scale_link,
            }
        )

    def theta_to_scipy_params(self, theta: np.ndarray) -> dict:
        mu = theta[:, 0]
        sigma = theta[:, 1]
        return {"s": sigma, "scale": robust_exp(mu), "loc": 0}

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma = self.theta_to_params(theta)

        if param == 0:
            return (np.log(y) - mu) / (sigma**2)

        if param == 1:
            return ((np.log(y) - mu) ** 2 - sigma**2) / (sigma**3)

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma = self.theta_to_params(theta)
        if param == 0:
            return -1 / (sigma**2)

        if param == 1:
            return -2 / (sigma**2)

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        if sorted(params) == [0, 1]:
            return np.zeros_like(y)

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        log_y = np.log(y)
        initial_params = [np.mean(log_y), np.std(log_y, ddof=1)]
        return np.tile(initial_params, (y.shape[0], 1))

    def pdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        """
        Probability density function of the Log-Normal distribution.
        """
        mu, sigma = self.theta_to_params(theta)
        return (
            1
            / (y * sigma * np.sqrt(2 * np.pi))
            * np.exp(-((np.log(y) - mu) ** 2) / (2 * sigma**2))
        )

    def cdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        """
        Cumulative distribution function of the Log-Normal distribution.
        """
        mu, sigma = self.theta_to_params(theta)
        return st.norm.cdf((np.log(y) - mu) / sigma)

    def ppf(self, p: np.ndarray, theta: np.ndarray) -> np.ndarray:
        """
        Percent-point function (quantile function) of the Log-Normal distribution.
        """
        mu, sigma = self.theta_to_params(theta)
        return np.exp(mu + sigma * st.norm.ppf(p))

    def logpdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        """
        Logarithm of the probability density function of the Log-Normal distribution.
        """
        mu, sigma = self.theta_to_params(theta)
        return np.log(1 / (y * sigma * np.sqrt(2 * np.pi))) + (
            -((np.log(y) - mu) ** 2) / (2 * sigma**2)
        )

    def logcdf(self, y, theta):
        return np.log(self.cdf(y, theta))



import numpy as np
import pytest
import scipy.stats as st
from ondil.gram import (
    init_gram,
    init_inverted_gram,
    init_y_gram,
    update_gram,
    update_inverted_gram,
    update_y_gram,
)


def make_x_y_w(N, D, random_weights=True):
    X = st.multivariate_normal().rvs((N, D))
    if D == 1:
        X = X.reshape(-1, 1)
    y = st.multivariate_normal().rvs((N, 1))
    if random_weights:
        w = st.uniform().rvs(N)
    else:
        w = np.ones(N)
    return X, y, w


N = [100, 1000]
D = [1, 2, 10]
RANDOM_WEIGHTS = [True, False]
FORGET = [0, 0.0001, 0.001, 0.01, 0.1]
BATCH_SIZE = [10, 25]


@pytest.mark.parametrize("N", N, ids=lambda x: f"N_{x}")
@pytest.mark.parametrize("D", D, ids=lambda x: f"D_{x}")
@pytest.mark.parametrize(
    "random_weights", RANDOM_WEIGHTS, ids=lambda x: f"random_weights_{x}"
)
@pytest.mark.parametrize("forget", FORGET, ids=lambda x: f"forget_{x}")
def test_inverse_rank_deficit(N, D, random_weights, forget):
    X, _, w = make_x_y_w(N, D, random_weights=random_weights)
    for d in range(1, D + 1):
        choice = np.random.choice(np.arange(D), d)
        XX = np.hstack((X, X[:, choice]))
        with pytest.raises(ValueError):
            _ = init_inverted_gram(XX[:-1], w[:-1], forget)


@pytest.mark.parametrize("N", N, ids=lambda x: f"N_{x}")
@pytest.mark.parametrize("D", D, ids=lambda x: f"D_{x}")
@pytest.mark.parametrize(
    "random_weights", RANDOM_WEIGHTS, ids=lambda x: f"random_weights_{x}"
)
@pytest.mark.parametrize("forget", FORGET, ids=lambda x: f"forget_{x}")
def test_single_update_x_gram(N, D, random_weights, forget):
    X, _, w = make_x_y_w(N, D, random_weights=random_weights)
    gram_start = init_gram(X[:-1], w[:-1], forget)
    gram_final = init_gram(X, w, forget)
    assert np.allclose(
        gram_final, update_gram(gram_start, X[[-1]], forget=forget, w=w[-1])
    )


@pytest.mark.parametrize("N", N, ids=lambda x: f"N_{x}")
@pytest.mark.parametrize("D", D, ids=lambda x: f"D_{x}")
@pytest.mark.parametrize(
    "random_weights", RANDOM_WEIGHTS, ids=lambda x: f"random_weights_{x}"
)
@pytest.mark.parametrize("forget", FORGET, ids=lambda x: f"forget_{x}")
@pytest.mark.parametrize("batchsize", BATCH_SIZE, ids=lambda x: f"batchsize_{x}")
def test_batch_update_x_gram(N, D, random_weights, forget, batchsize):
    X, _, w = make_x_y_w(N, D, random_weights=random_weights)
    gram_start = init_gram(X[:-batchsize], w[:-batchsize], forget)
    gram_final = init_gram(X, w, forget)
    assert np.allclose(
        gram_final,
        update_gram(gram_start, X[-batchsize:, :], forget=forget, w=w[-batchsize:]),
    )


# INVERTED GRAM
@pytest.mark.parametrize("N", N, ids=lambda x: f"N_{x}")
@pytest.mark.parametrize("D", D, ids=lambda x: f"D_{x}")
@pytest.mark.parametrize(
    "random_weights", RANDOM_WEIGHTS, ids=lambda x: f"random_weights_{x}"
)
@pytest.mark.parametrize("forget", FORGET, ids=lambda x: f"forget_{x}")
def test_single_update_inv_gram(N, D, random_weights, forget):
    X, _, w = make_x_y_w(N, D, random_weights=random_weights)
    gram_start = init_inverted_gram(X[:-1], w[:-1], forget)
    gram_final = init_inverted_gram(X, w, forget)
    assert np.allclose(
        gram_final, update_inverted_gram(gram_start, X[[-1]], forget=forget, w=w[-1])
    )


@pytest.mark.parametrize("N", N, ids=lambda x: f"N_{x}")
@pytest.mark.parametrize("D", D, ids=lambda x: f"D_{x}")
@pytest.mark.parametrize(
    "random_weights", RANDOM_WEIGHTS, ids=lambda x: f"random_weights_{x}"
)
@pytest.mark.parametrize("forget", FORGET, ids=lambda x: f"forget_{x}")
@pytest.mark.parametrize("batchsize", BATCH_SIZE, ids=lambda x: f"batchsize_{x}")
def test_batch_update_inv_gram(N, D, random_weights, forget, batchsize):
    X, _, w = make_x_y_w(N, D, random_weights=random_weights)
    gram_start = init_inverted_gram(X[:-batchsize], w[:-batchsize], forget)
    gram_final = init_inverted_gram(X, w, forget)
    assert np.allclose(
        gram_final,
        update_inverted_gram(
            gram_start, X[-batchsize:, :], forget=forget, w=w[-batchsize:]
        ),
    )


# Y-GRAM
@pytest.mark.parametrize("N", N, ids=lambda x: f"N_{x}")
@pytest.mark.parametrize("D", D, ids=lambda x: f"D_{x}")
@pytest.mark.parametrize(
    "random_weights", RANDOM_WEIGHTS, ids=lambda x: f"random_weights_{x}"
)
@pytest.mark.parametrize("forget", FORGET, ids=lambda x: f"forget_{x}")
def test_single_update_y_gram(N, D, random_weights, forget):
    X, y, w = make_x_y_w(N, D, random_weights=random_weights)
    gram_start = init_y_gram(X[:-1], y[:-1], w[:-1], forget)
    gram_final = init_y_gram(X, y, w, forget)
    assert np.allclose(
        gram_final, update_y_gram(gram_start, X[[-1]], y[[-1]], forget=forget, w=w[-1])
    )


@pytest.mark.parametrize("N", N, ids=lambda x: f"N_{x}")
@pytest.mark.parametrize("D", D, ids=lambda x: f"D_{x}")
@pytest.mark.parametrize(
    "random_weights", RANDOM_WEIGHTS, ids=lambda x: f"random_weights_{x}"
)
@pytest.mark.parametrize("forget", FORGET, ids=lambda x: f"forget_{x}")
@pytest.mark.parametrize("batchsize", BATCH_SIZE, ids=lambda x: f"batchsize_{x}")
def test_batch_update_y_gram(N, D, random_weights, forget, batchsize):
    X, y, w = make_x_y_w(N, D, random_weights=random_weights)
    gram_start = init_y_gram(X[:-batchsize], y[:-batchsize], w[:-batchsize], forget)
    gram_final = init_y_gram(X, y, w, forget)
    assert np.allclose(
        gram_final,
        update_y_gram(
            gram_start,
            X[-batchsize:, :],
            y[-batchsize:],
            forget=forget,
            w=w[-batchsize:],
        ),
    )



from typing import Dict, Optional, Tuple

import numpy as np
import scipy.stats as st
from scipy.special import digamma, polygamma

from ..base import Distribution, LinkFunction, ScipyMixin
from ..links import Log


class InverseGamma(ScipyMixin, Distribution):
    """
    The Inverse Gamma distribution as parameterized in GAMLSS:

    Parameters:
        - mu: mean-related parameter
        - sigma: dispersion parameter

    Reparameterization:
        α = 1 / sigma²
        scale = mu * (1 + sigma²) / sigma²

    This distribution corresponds to IGAMMA() in GAMLSS.
    """

    corresponding_gamlss: str = "IGAMMA"
    parameter_names = {0: "mu", 1: "sigma"}
    parameter_support = {
        0: (np.nextafter(0, 1), np.inf),
        1: (np.nextafter(0, 1), np.inf),
    }
    distribution_support = (0, np.inf)
    scipy_dist = st.invgamma
    scipy_names = {"mu": "scale", "sigma": "a"}

    def __init__(
        self,
        loc_link: LinkFunction = Log(),
        scale_link: LinkFunction = Log(),
    ) -> None:
        super().__init__(
            links={
                0: loc_link,
                1: scale_link,
            }
        )

    def theta_to_scipy_params(self, theta: np.ndarray) -> Dict[str, np.ndarray]:
        mu, sigma = self.theta_to_params(theta)
        alpha = 1.0 / sigma**2
        scale = mu * (1 + sigma**2) / sigma**2
        return {"a": alpha, "scale": scale, "loc": 0}

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma = self.theta_to_params(theta)
        alpha = 1.0 / sigma**2

        if param == 0:
            return (alpha / mu) - ((alpha + 1) / y)
        if param == 1:
            term = (
                np.log(mu)
                + (alpha / (alpha + 1))
                + np.log(alpha + 1)
                - digamma(alpha)
                - np.log(y)
                - (mu / y)
            )
            return (-2 / sigma**3) * term

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma = self.theta_to_params(theta)
        alpha = 1.0 / sigma**2

        if param == 0:
            return -1.0 / (sigma**2 * mu**2)
        elif param == 1:
            part1 = (sigma**2 * (1 + 2 * sigma**2)) / ((1 + sigma**2) ** 2)
            part2 = polygamma(1, alpha)
            return -4 * (part2 - part1) / (sigma**6)
        else:
            raise ValueError(f"Invalid parameter index: {param}")

    def sample(self, n: int, theta: np.ndarray) -> np.ndarray:
        mu, sigma = self.theta_to_params(theta)
        alpha = 1.0 / sigma**2
        scale = mu * (1 + sigma**2) / sigma**2
        return st.invgamma.rvs(a=alpha, scale=scale, size=n)

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        mu, sigma = self.theta_to_params(theta)
        if sorted(params) == [0, 1]:
            return -(2 / (mu * sigma * (1 + sigma**2)))

    def initial_values(
        self, y: np.ndarray, param: Optional[int] = None, axis: Optional[int] = None
    ) -> np.ndarray:
        y = np.asarray(y)
        mu_init = np.mean(y, axis=axis)
        sigma_init = np.std(y, ddof=1) / np.sqrt((mu_init))
        if param == 0:
            return np.full_like(y, mu_init)
        elif param == 1:
            return np.full_like(y, sigma_init)
        initial_params = [mu_init, sigma_init]
        return np.tile(initial_params, (y.shape[0], 1))



from typing import Tuple

import numpy as np
import scipy.stats as st

from ..base import Distribution, LinkFunction, ScipyMixin
from ..links import Identity, Log


class JSU(ScipyMixin, Distribution):
    """
    Corresponds to GAMLSS JSUo() and scipy.stats.johnsonsu()

    Distribution parameters:
    0 : Location
    1 : Scale (close to standard deviation)
    2 : Skewness
    3 : Tail behaviour
    """

    corresponding_gamlss: str = "JSUo"

    parameter_names = {0: "mu", 1: "sigma", 2: "nu", 3: "tau"}
    parameter_support = {
        0: (-np.inf, np.inf),
        1: (np.nextafter(0, 1), np.inf),
        2: (-np.inf, np.inf),
        3: (np.nextafter(0, 1), np.inf),
    }
    distribution_support = (-np.inf, np.inf)

    # Scipy equivalent and parameter mapping ondil -> scipy
    scipy_dist = st.johnsonsu
    scipy_names = {"mu": "loc", "sigma": "scale", "nu": "a", "tau": "b"}

    def __init__(
        self,
        loc_link: LinkFunction = Identity(),
        scale_link: LinkFunction = Log(),
        skew_link: LinkFunction = Identity(),
        tail_link: LinkFunction = Log(),
        use_gamlss_init_values: bool = False,
    ) -> None:
        super().__init__(
            links={
                0: loc_link,
                1: scale_link,
                2: skew_link,
                3: tail_link,
            }
        )
        self.gamlss_init_values = use_gamlss_init_values

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma, nu, tau = self.theta_to_params(theta)

        if param == 0:
            # MU
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldm = (z / (sigma * (z * z + 1))) + (
                (r * tau) / (sigma * np.sqrt(z * z + 1))
            )
            return dldm

        if param == 1:
            # SIGMA
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldd = (-1 / (sigma * (z * z + 1))) + (
                (r * tau * z) / (sigma * np.sqrt(z * z + 1))
            )
            return dldd

        if param == 2:
            # nu
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldv = -r
            return dldv

        if param == 3:
            # tau
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldt = (1 + r * nu - r * r) / tau
            return dldt

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma, nu, tau = self.theta_to_params(theta)
        if param == 0:
            # MU
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldm = (z / (sigma * (z * z + 1))) + (
                (r * tau) / (sigma * np.sqrt(z * z + 1))
            )
            d2ldm2 = -dldm * dldm
            return d2ldm2

        if param == 1:
            # SIGMA
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldd = (-1 / (sigma * (z * z + 1))) + (
                (r * tau * z) / (sigma * np.sqrt(z * z + 1))
            )
            d2ldd2 = -(dldd * dldd)
            return d2ldd2

        if param == 2:
            # TAIL
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            d2ldv2 = -(r * r)
            return d2ldv2

        if param == 3:
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldt = (1 + r * nu - r * r) / tau
            d2ldt2 = -dldt * dldt
            return d2ldt2

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        mu, sigma, nu, tau = self.theta_to_params(theta)
        if sorted(params) == [0, 1]:
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldm = (z / (sigma * (z * z + 1))) + (
                (r * tau) / (sigma * np.sqrt(z * z + 1))
            )
            dldd = (-1 / (sigma * (z * z + 1))) + (
                (r * tau * z) / (sigma * np.sqrt(z * z + 1))
            )
            d2ldmdd = -(dldm * dldd)
            return d2ldmdd

        if sorted(params) == [0, 2]:
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldm = (z / (sigma * (z * z + 1))) + (
                (r * tau) / (sigma * np.sqrt(z * z + 1))
            )
            dldv = -r
            d2ldmdv = -(dldm * dldv)
            return d2ldmdv

        if sorted(params) == [0, 3]:
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldm = (z / (sigma * (z * z + 1))) + (
                (r * tau) / (sigma * np.sqrt(z * z + 1))
            )
            dldt = (1 + r * nu - r * r) / tau
            d2ldmdt = -(dldm * dldt)
            return d2ldmdt

        if sorted(params) == [1, 2]:
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldd = (-1 / (sigma * (z * z + 1))) + (
                (r * tau * z) / (sigma * np.sqrt(z * z + 1))
            )
            dldv = -r
            d2ldddv = -(dldd * dldv)
            return d2ldddv

        if sorted(params) == [1, 3]:
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldd = (-1 / (sigma * (z * z + 1))) + (
                (r * tau * z) / (sigma * np.sqrt(z * z + 1))
            )
            dldt = (1 + r * nu - r * r) / tau
            d2ldddt = -(dldd * dldt)
            return d2ldddt

        if sorted(params) == [2, 3]:
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldv = -r
            dldt = (1 + r * nu - r * r) / tau
            d2ldvdt = -(dldv * dldt)
            return d2ldvdt

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        out = np.empty((y.shape[0], self.n_params))
        if self.gamlss_init_values:
            out[:, 0] = (np.repeat(np.mean(y), y.shape[0]) + y) / 2
            out[:, 1] = 0.1
            out[:, 2] = 0.0
            out[:, 3] = 0.5
        else:
            params = st.johnsonsu.fit(y)
            out[:, 0] = params[2]
            out[:, 1] = params[3]
            out[:, 2] = params[0]
            out[:, 3] = params[1]
        return out



from typing import Literal

import numpy as np

from ..estimators.online_linear_model import OnlineLinearModel
from ..methods import LassoPath


class OnlineLasso(OnlineLinearModel):
    def __init__(
        self,
        forget: float = 0,
        scale_inputs: bool = True,
        fit_intercept: bool = True,
        regularize_intercept: bool = False,
        ic: Literal["aic", "bic", "hqc", "max"] = "bic",
        early_stop: int = 0,
        beta_lower_bound: np.ndarray | None = None,
        beta_upper_bound: np.ndarray | None = None,
        lambda_n: int = 100,
        lambda_eps: float = 1e-4,
        start_value: str = "previous_fit",
        tolerance: float = 1e-4,
        max_iterations: int = 1000,
        selection: Literal["cyclic", "random"] = "cyclic",
    ):
        """Online LASSO estimator class.

        This class initializes the online linear regression fitted using LASSO. The estimator object provides three main methods,
        ``estimator.fit(X, y)``, ``estimator.update(X, y)`` and ``estimator.predict(X)``.

        Args:
            forget (float, optional): Exponential discounting of old observations. Defaults to 0.
            scale_inputs (bool, optional): Whether to scale the $X$ matrix. Defaults to True.
            fit_intercept (bool, optional): Whether to add an intercept in the estimation. Defaults to True.
            regularize_intercept (bool, optional): Whether to regularize the intercept. Defaults to False.
            ic (Literal["aic", "bic", "hqc", "max"], optional): The information criteria for model selection. Defaults to "bic".
            early_stop (int, optional): Early stopping criterion. If we reach `early_stop` non-zero coefficients, we stop. Defaults to 0 (no early stopping).
            beta_lower_bound (np.ndarray | None, optional): Lower bounds for beta. Keep in mind the size of X and whether you want to fit an intercept. None corresponds to unconstrained estimation.Defaults to None.
            beta_upper_bound (np.ndarray | None, optional): Lower bounds for beta. Keep in mind the size of X and whether you want to fit an intercept. None corresponds to unconstrained estimation. Defaults to None.
            lambda_n (int, optional): Length of the regularization path. Defaults to 100.
            lambda_eps (float, optional): The largest regularization is determined automatically such that the solution is fully regularized. The smallest regularization is taken as $\\varepsilon  \\lambda^\max$ and we will use an exponential grid. Defaults to 1e-4.
            start_value (str, optional): Whether to choose the previous fit or the previous regularization as start value. Defaults to 100.
            tolerance (float, optional): Tolerance for breaking the CD. Defaults to 1e-4.
            max_iterations (int, optional): Max number of CD iterations. Defaults to 1000.
            selection (Literal["cyclic", "random"], optional): Whether to cycle through all coordinates in order or random. For large problems, random might increase convergence. Defaults to 100.
        """

        concrete_method = LassoPath(
            lambda_eps=lambda_eps,
            lambda_n=lambda_n,
            start_value_initial="previous_lambda",
            early_stop=early_stop,
            start_value_update=start_value,
            tolerance=tolerance,
            max_iterations=max_iterations,
            selection=selection,
            beta_lower_bound=beta_lower_bound,
            beta_upper_bound=beta_upper_bound,
        )
        super().__init__(
            forget=forget,
            scale_inputs=scale_inputs,
            fit_intercept=fit_intercept,
            regularize_intercept=regularize_intercept,
            method=concrete_method,
            ic=ic,
        )



from typing import Literal, Tuple

import numba as nb
import numpy as np


@nb.njit()
def soft_threshold(value: float, threshold: float):
    """The soft thresholding function.

    For value \(x\) and threshold \(\\lambda\), the soft thresholding function \(S(x, \\lambda)\) is
    defined as:

    $$S(x, \\lambda) = sign(x)(|x| - \\lambda)$$

    Args:
        value (float): The value
        threshold (float): The threshold

    Returns:
        out (float): The thresholded value
    """
    return np.sign(value) * np.maximum(np.abs(value) - threshold, 0)


# @nb.njit([
#     "(float64[:,:], float64[:], float64[:], float64, bool[:], str, float64, int64)",
#    "(float32[:,:], float32[:], float32[:], float32, bool[:], str, float32, int32)",
# ])
@nb.njit()
def online_coordinate_descent(
    x_gram: np.ndarray,
    y_gram: np.ndarray,
    beta: np.ndarray,
    regularization: float,
    is_regularized: bool,
    alpha: float,
    beta_lower_bound: np.ndarray,
    beta_upper_bound: np.ndarray,
    selection: Literal["cyclic", "random"] = "cyclic",
    tolerance: float = 1e-4,
    max_iterations: int = 1000,
) -> Tuple[np.ndarray, int]:
    """The parameter update cycle of the online coordinate descent.

    Args:
        x_gram (np.ndarray): X-Gramian $$X^TX$$
        y_gram (np.ndarray): Y-Gramian $$X^TY$$
        beta (np.ndarray): Current beta vector
        regularization (float): Regularization parameter lambda
        is_regularized (bool): Vector of bools indicating whether the coefficient is regularized
        beta_lower_bound (np.ndarray): Lower bounds for beta
        beta_upper_bound (np.ndarray): Upper bounds for beta
        selection (Literal['cyclic', 'random'], optional): Apply cyclic or random coordinate descent. Defaults to "cyclic".
        tolerance (float, optional): Tolerance for the beta update. Defaults to 1e-4.
        max_iterations (int, optional): Maximum iterations. Defaults to 1000.

    Returns:
        Tuple[np.ndarray, int]: Converged $$ \\beta $$
    """
    i = 0
    J = beta.shape[0]
    JJ = np.arange(J)
    beta_now = np.copy(beta)
    beta_star = np.copy(beta)

    while True:
        i += 1
        beta_star = np.copy(beta_now)

        if (selection == "random") and (i >= 2):
            JJ = np.random.permutation(J)

        for j in JJ:
            if (i < 2) | (beta_now[j] != 0):
                update = (
                    y_gram[j] - (x_gram[j, :] @ beta_now) + x_gram[j, j] * beta_now[j]
                )
                if is_regularized[j]:
                    update = soft_threshold(update, alpha * regularization)
                    denom = x_gram[j, j] + regularization * (1 - alpha)
                else:
                    denom = x_gram[j, j]
                beta_now[j] = min(
                    max(update / denom, beta_lower_bound[j]), beta_upper_bound[j]
                )
        if np.max(np.abs(beta_now - beta_star)) <= tolerance * np.max(np.abs(beta_now)):
            break
        if i > max_iterations:
            break
    return beta_now, i


# @nb.njit([
#     "(float64[:, :])(float64[:, :], float64[:], float64[:, :], float64[:], bool[:], str, float64, int64)",
#     "(float32[:, :])(float32[:, :], float32[:], float32[:, :], float32[:], bool[:], str, float32, int32)",
# ])
@nb.njit()
def online_coordinate_descent_path(
    x_gram: np.ndarray,
    y_gram: np.ndarray,
    beta_path: np.ndarray,
    lambda_path: np.ndarray,
    is_regularized: np.ndarray,
    alpha: float,
    early_stop: int,
    beta_lower_bound: np.ndarray,
    beta_upper_bound: np.ndarray,
    which_start_value: Literal[
        "previous_lambda", "previous_fit", "average"
    ] = "previous_lambda",
    selection: Literal["cyclic", "random"] = "cyclic",
    tolerance: float = 1e-4,
    max_iterations: int = 1000,
) -> Tuple[np.ndarray, np.ndarray]:
    """Run coordinate descent on a grid of regularization values.

    Args:
        x_gram (np.ndarray): X-Gramian $$X^TX$$
        y_gram (np.ndarray): Y-Gramian $$X^TY$$
        beta_path (np.ndarray): The current coefficent path
        lambda_path (np.ndarray): The lambda grid
        is_regularized (bool): Vector of bools indicating whether the coefficient is regularized
        early_stop (int, optional): Early stopping criterion. 0 implies no early stopping. Defaults to 0.
        beta_lower_bound (np.ndarray): Lower bounds for beta
        beta_upper_bound (np.ndarray): Upper bounds for beta
        which_start_value (Literal['previous_lambda', 'previous_fit', 'average'], optional): Values to warm-start the coordinate descent. Defaults to "previous_lambda".
        selection (Literal['cyclic', 'random'], optional): Apply cyclic or random coordinate descent. Defaults to "cyclic".
        tolerance (float, optional): Tolerance for the beta update. Will be passed through to the parameter update. Defaults to 1e-4.
        max_iterations (int, optional): Maximum iterations. Will be passed through to the parameter update. Defaults to 1000.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Tuple with the updated coefficient path and the iteration count.
    """

    beta_path_new = np.zeros_like(beta_path)
    iterations = np.zeros_like(lambda_path)

    for i, regularization in enumerate(lambda_path):
        # Select the according start values for the next CD update
        if which_start_value == "average":
            beta = (beta_path_new[max(i - 1, 0), :] + beta_path[max(i, 0), :]) / 2
        if which_start_value == "previous_lambda":
            beta = beta_path_new[max(i - 1, 0), :]
        else:
            beta = beta_path[max(i, 0), :]

        if (early_stop > 0) and np.count_nonzero(beta) >= early_stop:
            beta_path_new[i, :] = beta
            iterations[i] = 0
        else:
            beta_path_new[i, :], iterations[i] = online_coordinate_descent(
                x_gram=x_gram,
                y_gram=y_gram,
                beta=beta,
                regularization=regularization,
                is_regularized=is_regularized,
                alpha=alpha,
                beta_lower_bound=beta_lower_bound,
                beta_upper_bound=beta_upper_bound,
                selection=selection,
                tolerance=tolerance,
                max_iterations=max_iterations,
            )

    return beta_path_new, iterations



# Author: Simon Hirsch
# A real basic test to check that the import does not fail.


def test_import():
    try:
        import ondil  # noqa

        failed = False
    except Exception:
        failed = True

    assert not failed, "Import failed with Exception."
